#include "stdio.h"
#include "netinet/in.h"
#include "net/if.h"
#include "sys/ioctl.h"
#include "net/if_arp.h"
#include "string.h"

/* obtained from redboot.cdl */
#define MAX_SCRIPT_LENGTH  (512)
#define MAX_STRING_LENGTH  (128)
#define MAX_CONFIG_DATA    (4096)

#define CONFIG_EMPTY   0
#define CONFIG_BOOL    1
#define CONFIG_INT     2
#define CONFIG_STRING  3
#define CONFIG_SCRIPT  4
#define CONFIG_IP      5
#define CONFIG_ESA     6
#define CONFIG_NETPORT 7

#define CONFIG_OBJECT_TYPE(dp)          (dp)[0]
#define CONFIG_OBJECT_KEYLEN(dp)        (dp)[1]
#define CONFIG_OBJECT_ENABLE_SENSE(dp)  (dp)[2]
#define CONFIG_OBJECT_ENABLE_KEYLEN(dp) (dp)[3]
#define CONFIG_OBJECT_KEY(dp)           ((dp)+4)
#define CONFIG_OBJECT_ENABLE_KEY(dp)    ((dp)+4+CONFIG_OBJECT_KEYLEN(dp))
#define CONFIG_OBJECT_VALUE(dp)         ((dp)+4+CONFIG_OBJECT_KEYLEN(dp)+CONFIG_OBJECT_ENABLE_KEYLEN(dp))


static struct _config {
    unsigned long len;
    unsigned long key1;
    unsigned char config_data[MAX_CONFIG_DATA-(4*4)];
    unsigned long key2;
    unsigned long cksum;
} config;

static int
config_length(int type)
{
    switch (type) {
    case CONFIG_BOOL:
        return sizeof(int);
    case CONFIG_INT:
        return sizeof(unsigned long);
    case CONFIG_IP:
        return sizeof(int);
    case CONFIG_ESA:
        // Would like this to be sizeof(enet_addr_t), but that causes much
        // pain since it fouls the alignment of data which follows.
        return 8;
    case CONFIG_NETPORT:
        return MAX_STRING_LENGTH;
    case CONFIG_STRING:
        return MAX_STRING_LENGTH;
    case CONFIG_SCRIPT:
        return MAX_SCRIPT_LENGTH;
    default:
        return 0;
    }
}

static unsigned char *
flash_lookup_config(char *key)
{
    unsigned char *dp;
    int len;

    dp = &config.config_data[0];
    while (dp < &config.config_data[sizeof(config.config_data)]) {
        len = 4 + CONFIG_OBJECT_KEYLEN(dp) + CONFIG_OBJECT_ENABLE_KEYLEN(dp) +
            config_length(CONFIG_OBJECT_TYPE(dp));
        if (strcmp(key, CONFIG_OBJECT_KEY(dp)) == 0) {
            return dp;
        }
        dp += len;
    }
    return NULL;
}

int main(int argc, char ** argv)
    {
    int result = 0;
    unsigned char * cfg = NULL;
    FILE * fp = NULL;
    struct ifreq ifr;
    int s = -1;
    struct sockaddr * sa;
    int verbose = 0;
    if (argv[1] != NULL && argv[1][0] == 'v')
        verbose = 1;

    fp = fopen("/dev/rbcfgB", "r");
    if (fp == NULL)
        {
        perror("fopen");
        exit(1);
        }

    result = fread(&config, 1, sizeof(config), fp); 
    if (result != sizeof(config))
        {
        perror("fread");
        exit(1);
        }

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s == -1)
        {
        perror("socket");
        exit(1);
        }

    cfg = flash_lookup_config("npe_eth0_esa");
    if (cfg != NULL && CONFIG_OBJECT_TYPE(cfg) == CONFIG_ESA)
        {
        cfg = CONFIG_OBJECT_VALUE(cfg);
        if (verbose)
            printf("%s MAC: %x:%x:%x:%x:%x:%x\n", "ixp0",
                   cfg[0], cfg[1], cfg[2], cfg[3], cfg[4], cfg[5]);
        memset(&ifr, 0, sizeof(ifr));
        strcpy(ifr.ifr_name, "ixp0");
        sa = (struct sockaddr *)&ifr.ifr_hwaddr;
        sa->sa_family = ARPHRD_ETHER;
        memcpy(sa->sa_data, cfg, 6);
        if(ioctl(s, SIOCSIFHWADDR, &ifr) < 0)
            perror("ioctl");
        }
    cfg = flash_lookup_config("npe_eth1_esa");
    if (cfg != NULL && CONFIG_OBJECT_TYPE(cfg) == CONFIG_ESA)
        {
        cfg = CONFIG_OBJECT_VALUE(cfg);
        if (verbose)
            printf("%s MAC: %x:%x:%x:%x:%x:%x\n", "ixp1",
                   cfg[0], cfg[1], cfg[2], cfg[3], cfg[4], cfg[5]);
        memset(&ifr, 0, sizeof(ifr));
        strcpy(ifr.ifr_name, "ixp1");
        sa->sa_family = ARPHRD_ETHER;
        memcpy(sa->sa_data, cfg, 6);
        if(ioctl(s, SIOCSIFHWADDR, &ifr) < 0)
            perror("ioctl");
        }
    cfg = flash_lookup_config("rtl_8169_esa");
    if (cfg != NULL && CONFIG_OBJECT_TYPE(cfg) == CONFIG_ESA)
        {
        cfg = CONFIG_OBJECT_VALUE(cfg);
        if (verbose)
            printf("%s MAC: %x:%x:%x:%x:%x:%x\n", "eth0",
                   cfg[0], cfg[1], cfg[2], cfg[3], cfg[4], cfg[5]);
        memset(&ifr, 0, sizeof(ifr));
        strcpy(ifr.ifr_name, "eth0");
        sa->sa_family = ARPHRD_ETHER;
        memcpy(sa->sa_data, cfg, 6);
        if(ioctl(s, SIOCSIFHWADDR, &ifr) < 0)
            perror("ioctl");
        }

    return 0;
    }
